@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -26,37 +26,49 @@
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <errno.h>
 
 #include <glib.h>
 #include <gdbus.h>
 
 #include "ofono.h"
 
-#include "driver.h"
 #include "common.h"
-#include "ussd.h"
-
-#define CALL_FORWARDING_INTERFACE "org.ofono.CallForwarding"
 
 #define CALL_FORWARDING_FLAG_CACHED 0x1
 
 /* According to 27.007 Spec */
 #define DEFAULT_NO_REPLY_TIMEOUT 20
 
-struct call_forwarding_data {
-	struct ofono_call_forwarding_ops *ops;
+static GSList *g_drivers = NULL;
+
+enum call_forwarding_type {
+	CALL_FORWARDING_TYPE_UNCONDITIONAL = 0,
+	CALL_FORWARDING_TYPE_BUSY = 1,
+	CALL_FORWARDING_TYPE_NO_REPLY = 2,
+	CALL_FORWARDING_TYPE_NOT_REACHABLE = 3,
+	CALL_FORWARDING_TYPE_ALL = 4,
+	CALL_FORWARDING_TYPE_ALL_CONDITIONAL = 5
+};
+
+struct ofono_call_forwarding {
 	GSList *cf_conditions[4];
 	int flags;
 	DBusMessage *pending;
 	int query_next;
 	int query_end;
 	struct cf_ss_request *ss_req;
+	struct ofono_ussd *ussd;
+	unsigned int ussd_watch;
+	const struct ofono_call_forwarding_driver *driver;
+	void *driver_data;
+	struct ofono_atom *atom;
 };
 
-static void get_query_next_cf_cond(struct ofono_modem *modem);
-static void set_query_next_cf_cond(struct ofono_modem *modem);
-static void ss_set_query_next_cf_cond(struct ofono_modem *modem);
-static void cf_unregister_ss_controls(struct ofono_modem *modem);
+static void get_query_next_cf_cond(struct ofono_call_forwarding *cf);
+static void set_query_next_cf_cond(struct ofono_call_forwarding *cf);
+static void ss_set_query_next_cf_cond(struct ofono_call_forwarding *cf);
+static void cf_unregister_ss_controls(struct ofono_call_forwarding *cf);
 
 struct cf_ss_request {
 	int ss_type;
@@ -67,8 +79,8 @@ struct cf_ss_request {
 
 static gint cf_condition_compare(gconstpointer a, gconstpointer b)
 {
-	const struct ofono_cf_condition *ca = a;
-	const struct ofono_cf_condition *cb = b;
+	const struct ofono_call_forwarding_condition *ca = a;
+	const struct ofono_call_forwarding_condition *cb = b;
 
 	if (ca->cls < cb->cls)
 		return -1;
@@ -81,7 +93,7 @@ static gint cf_condition_compare(gconstpointer a, gconstpointer b)
 
 static gint cf_condition_find_with_cls(gconstpointer a, gconstpointer b)
 {
-	const struct ofono_cf_condition *c = a;
+	const struct ofono_call_forwarding_condition *c = a;
 	int cls = GPOINTER_TO_INT(b);
 
 	if (c->cls < cls)
@@ -96,7 +108,7 @@ static gint cf_condition_find_with_cls(gconstpointer a, gconstpointer b)
 static int cf_find_timeout(GSList *cf_list, int cls)
 {
 	GSList *l;
-	struct ofono_cf_condition *c;
+	struct ofono_call_forwarding_condition *c;
 
 	l = g_slist_find_custom(cf_list, GINT_TO_POINTER(cls),
 		cf_condition_find_with_cls);
@@ -112,12 +124,12 @@ static int cf_find_timeout(GSList *cf_list, int cls)
 static void cf_cond_list_print(GSList *list)
 {
 	GSList *l;
-	struct ofono_cf_condition *cond;
+	struct ofono_call_forwarding_condition *cond;
 
 	for (l = list; l; l = l->next) {
 		cond = l->data;
 
-		ofono_debug("CF Condition status: %d, class: %d, number: %s,"
+		DBG("CF Condition status: %d, class: %d, number: %s,"
 			" number_type: %d, time: %d",
 			cond->status, cond->cls, cond->phone_number.number,
 			cond->phone_number.type, cond->time);
@@ -125,14 +137,15 @@ static void cf_cond_list_print(GSList *list)
 }
 
 static GSList *cf_cond_list_create(int total,
-					const struct ofono_cf_condition *list)
+			const struct ofono_call_forwarding_condition *list)
 {
 	GSList *l = NULL;
 	int i;
 	int j;
-	struct ofono_cf_condition *cond;
+	struct ofono_call_forwarding_condition *cond;
 
-	/* Specification is not really clear how the results are reported,
+	/*
+	 * Specification is not really clear how the results are reported,
 	 * so assume both multiple list items & compound values of class
 	 * are possible
 	 */
@@ -144,11 +157,12 @@ static GSList *cf_cond_list_create(int total,
 			if (list[i].status == 0)
 				continue;
 
-			cond = g_try_new0(struct ofono_cf_condition, 1);
+			cond = g_try_new0(struct ofono_call_forwarding_condition, 1);
 			if (!cond)
 				continue;
 
-			memcpy(cond, &list[i], sizeof(struct ofono_cf_condition));
+			memcpy(cond, &list[i],
+				sizeof(struct ofono_call_forwarding_condition));
 			cond->cls = j;
 
 			l = g_slist_insert_sorted(l, cond,
@@ -169,7 +183,7 @@ static inline void cf_list_clear(GSList *cf_list)
 	g_slist_free(cf_list);
 }
 
-static inline void cf_clear_all(struct call_forwarding_data *cf)
+static inline void cf_clear_all(struct ofono_call_forwarding *cf)
 {
 	int i;
 
@@ -179,30 +193,6 @@ static inline void cf_clear_all(struct call_forwarding_data *cf)
 	}
 }
 
-static struct call_forwarding_data *call_forwarding_create()
-{
-	struct call_forwarding_data *r;
-
-	r = g_try_new0(struct call_forwarding_data, 1);
-
-	if (!r)
-		return r;
-
-	return r;
-}
-
-static void call_forwarding_destroy(gpointer data)
-{
-	struct ofono_modem *modem = data;
-	struct call_forwarding_data *cf = modem->call_forwarding;
-
-	cf_clear_all(cf);
-
-	cf_unregister_ss_controls(modem);
-
-	g_free(cf);
-}
-
 static const char *cf_type_lut[] = {
 	"Unconditional",
 	"Busy",
@@ -212,15 +202,16 @@ static const char *cf_type_lut[] = {
 	"AllConditional"
 };
 
-static void set_new_cond_list(struct ofono_modem *modem, int type, GSList *list)
+static void set_new_cond_list(struct ofono_call_forwarding *cf,
+				int type, GSList *list)
 {
-	struct call_forwarding_data *cf = modem->call_forwarding;
 	GSList *old = cf->cf_conditions[type];
 	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(cf->atom);
 	GSList *l;
 	GSList *o;
-	struct ofono_cf_condition *lc;
-	struct ofono_cf_condition *oc;
+	struct ofono_call_forwarding_condition *lc;
+	struct ofono_call_forwarding_condition *oc;
 	const char *number;
 	dbus_uint16_t timeout;
 	char attr[64];
@@ -229,7 +220,8 @@ static void set_new_cond_list(struct ofono_modem *modem, int type, GSList *list)
 	for (l = list; l; l = l->next) {
 		lc = l->data;
 
-		/* New condition lists might have attributes we don't care about
+		/*
+		 * New condition lists might have attributes we don't care about
 		 * triggered by e.g. ss control magic strings just skip them
 		 * here.  For now we only support Voice, although Fax & all Data
 		 * basic services are applicable as well.
@@ -240,11 +232,11 @@ static void set_new_cond_list(struct ofono_modem *modem, int type, GSList *list)
 		timeout = lc->time;
 		number = phone_number_to_string(&lc->phone_number);
 
-		sprintf(attr, "%s%s", bearer_class_to_string(lc->cls),
-					cf_type_lut[type]);
+		snprintf(attr, sizeof(attr), "%s%s",
+			bearer_class_to_string(lc->cls), cf_type_lut[type]);
 
 		if (type == CALL_FORWARDING_TYPE_NO_REPLY)
-			sprintf(tattr, "%sTimeout", attr);
+			snprintf(tattr, sizeof(tattr), "%sTimeout", attr);
 
 		o = g_slist_find_custom(old, GINT_TO_POINTER(lc->cls),
 					cf_condition_find_with_cls);
@@ -255,17 +247,15 @@ static void set_new_cond_list(struct ofono_modem *modem, int type, GSList *list)
 			if (oc->phone_number.type != lc->phone_number.type ||
 				strcmp(oc->phone_number.number,
 					lc->phone_number.number))
-				ofono_dbus_signal_property_changed(conn,
-						modem->path,
-						CALL_FORWARDING_INTERFACE,
+				ofono_dbus_signal_property_changed(conn, path,
+						OFONO_CALL_FORWARDING_INTERFACE,
 						attr, DBUS_TYPE_STRING,
 						&number);
 
 			if (type == CALL_FORWARDING_TYPE_NO_REPLY &&
 				oc->time != lc->time)
-				ofono_dbus_signal_property_changed(conn,
-						modem->path,
-						CALL_FORWARDING_INTERFACE,
+				ofono_dbus_signal_property_changed(conn, path,
+						OFONO_CALL_FORWARDING_INTERFACE,
 						tattr, DBUS_TYPE_UINT16,
 						&timeout);
 
@@ -275,16 +265,15 @@ static void set_new_cond_list(struct ofono_modem *modem, int type, GSList *list)
 		} else {
 			number = phone_number_to_string(&lc->phone_number);
 
-			ofono_dbus_signal_property_changed(conn, modem->path,
-						CALL_FORWARDING_INTERFACE,
+			ofono_dbus_signal_property_changed(conn, path,
+						OFONO_CALL_FORWARDING_INTERFACE,
 						attr, DBUS_TYPE_STRING,
 						&number);
 
 			if (type == CALL_FORWARDING_TYPE_NO_REPLY &&
 				lc->time != DEFAULT_NO_REPLY_TIMEOUT)
-				ofono_dbus_signal_property_changed(conn,
-						modem->path,
-						CALL_FORWARDING_INTERFACE,
+				ofono_dbus_signal_property_changed(conn, path,
+						OFONO_CALL_FORWARDING_INTERFACE,
 						tattr, DBUS_TYPE_UINT16,
 						&timeout);
 		}
@@ -296,20 +285,20 @@ static void set_new_cond_list(struct ofono_modem *modem, int type, GSList *list)
 	for (o = old; o; o = o->next) {
 		oc = o->data;
 
-		sprintf(attr, "%s%s", bearer_class_to_string(oc->cls),
-					cf_type_lut[type]);
+		snprintf(attr, sizeof(attr), "%s%s",
+			bearer_class_to_string(oc->cls), cf_type_lut[type]);
 
 		if (type == CALL_FORWARDING_TYPE_NO_REPLY)
-			sprintf(tattr, "%sTimeout", attr);
+			snprintf(tattr, sizeof(tattr), "%sTimeout", attr);
 
-		ofono_dbus_signal_property_changed(conn, modem->path,
-					CALL_FORWARDING_INTERFACE, attr,
+		ofono_dbus_signal_property_changed(conn, path,
+					OFONO_CALL_FORWARDING_INTERFACE, attr,
 					DBUS_TYPE_STRING, &number);
 
 		if (type == CALL_FORWARDING_TYPE_NO_REPLY &&
 			oc->time != DEFAULT_NO_REPLY_TIMEOUT)
-			ofono_dbus_signal_property_changed(conn, modem->path,
-						CALL_FORWARDING_INTERFACE,
+			ofono_dbus_signal_property_changed(conn, path,
+						OFONO_CALL_FORWARDING_INTERFACE,
 						tattr, DBUS_TYPE_UINT16,
 						&timeout);
 	}
@@ -327,11 +316,12 @@ static inline void property_append_cf_condition(DBusMessageIter *dict, int cls,
 	char tattr[64];
 	int addt = !strcmp(postfix, "NoReply");
 
-	sprintf(attr, "%s%s", bearer_class_to_string(cls), postfix);
+	snprintf(attr, sizeof(attr), "%s%s",
+			bearer_class_to_string(cls), postfix);
 
 	if (addt)
-		sprintf(tattr, "%s%sTimeout", bearer_class_to_string(cls),
-			postfix);
+		snprintf(tattr, sizeof(tattr), "%s%sTimeout",
+				bearer_class_to_string(cls), postfix);
 
 	ofono_dbus_dict_append(dict, attr, DBUS_TYPE_STRING, &value);
 
@@ -345,7 +335,7 @@ static void property_append_cf_conditions(DBusMessageIter *dict,
 {
 	GSList *l;
 	int i;
-	struct ofono_cf_condition *cf;
+	struct ofono_call_forwarding_condition *cf;
 	const char *number;
 
 	for (i = 1, l = cf_list; i <= BEARER_CLASS_PAD; i = i << 1) {
@@ -369,7 +359,7 @@ static void property_append_cf_conditions(DBusMessageIter *dict,
 }
 
 static DBusMessage *cf_get_properties_reply(DBusMessage *msg,
-						struct call_forwarding_data *cf)
+						struct ofono_call_forwarding *cf)
 {
 	DBusMessage *reply;
 	DBusMessageIter iter;
@@ -398,18 +388,17 @@ static DBusMessage *cf_get_properties_reply(DBusMessage *msg,
 }
 
 static void get_query_cf_callback(const struct ofono_error *error, int total,
-					const struct ofono_cf_condition *list,
-					void *data)
+			const struct ofono_call_forwarding_condition *list,
+			void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_forwarding_data *cf = modem->call_forwarding;
+	struct ofono_call_forwarding *cf = data;
 
 	if (error->type == OFONO_ERROR_TYPE_NO_ERROR) {
 		GSList *l;
 		l = cf_cond_list_create(total, list);
-		set_new_cond_list(modem, cf->query_next, l);
+		set_new_cond_list(cf, cf->query_next, l);
 
-		ofono_debug("%s conditions:", cf_type_lut[cf->query_next]);
+		DBG("%s conditions:", cf_type_lut[cf->query_next]);
 		cf_cond_list_print(l);
 
 		if (cf->query_next == CALL_FORWARDING_TYPE_NOT_REACHABLE)
@@ -423,41 +412,39 @@ static void get_query_cf_callback(const struct ofono_error *error, int total,
 	}
 
 	cf->query_next++;
-	get_query_next_cf_cond(modem);
+	get_query_next_cf_cond(cf);
 }
 
-static void get_query_next_cf_cond(struct ofono_modem *modem)
+static void get_query_next_cf_cond(struct ofono_call_forwarding *cf)
 {
-	struct call_forwarding_data *cf = modem->call_forwarding;
-
-	cf->ops->query(modem, cf->query_next, BEARER_CLASS_DEFAULT,
-			get_query_cf_callback, modem);
+	cf->driver->query(cf, cf->query_next, BEARER_CLASS_DEFAULT,
+			get_query_cf_callback, cf);
 }
 
 static DBusMessage *cf_get_properties(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_forwarding_data *cf = modem->call_forwarding;
+	struct ofono_call_forwarding *cf = data;
 
 	if (cf->flags & CALL_FORWARDING_FLAG_CACHED)
 		return cf_get_properties_reply(msg, cf);
 
-	if (!cf->ops->query)
+	if (!cf->driver->query)
 		return __ofono_error_not_implemented(msg);
 
-	if (cf->pending)
+	if (__ofono_call_forwarding_is_busy(cf) ||
+			__ofono_ussd_is_busy(cf->ussd))
 		return __ofono_error_busy(msg);
 
 	cf->pending = dbus_message_ref(msg);
 	cf->query_next = 0;
 
-	get_query_next_cf_cond(modem);
+	get_query_next_cf_cond(cf);
 
 	return NULL;
 }
 
-static gboolean cf_condition_enabled_property(struct call_forwarding_data *cf,
+static gboolean cf_condition_enabled_property(struct ofono_call_forwarding *cf,
 			const char *property, int *out_type, int *out_cls)
 {
 	int i;
@@ -473,7 +460,8 @@ static gboolean cf_condition_enabled_property(struct call_forwarding_data *cf,
 		if (strncmp(property, prefix, len))
 			continue;
 
-		/* We check the 4 call forwarding types, e.g.
+		/*
+		 * We check the 4 call forwarding types, e.g.
 		 * unconditional, busy, no reply, not reachable
 		 */
 		for (j = 0; j < 4; j++)
@@ -512,11 +500,10 @@ static gboolean cf_condition_timeout_property(const char *property,
 }
 
 static void set_query_cf_callback(const struct ofono_error *error, int total,
-					const struct ofono_cf_condition *list,
-					void *data)
+			const struct ofono_call_forwarding_condition *list,
+			void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_forwarding_data *cf = modem->call_forwarding;
+	struct ofono_call_forwarding *cf = data;
 	GSList *l;
 	DBusMessage *reply;
 
@@ -531,70 +518,64 @@ static void set_query_cf_callback(const struct ofono_error *error, int total,
 	if (cf->query_next == cf->query_end) {
 		reply = dbus_message_new_method_return(cf->pending);
 		__ofono_dbus_pending_reply(&cf->pending, reply);
-	} 
+	}
 
 	l = cf_cond_list_create(total, list);
-	set_new_cond_list(modem, cf->query_next, l);
+	set_new_cond_list(cf, cf->query_next, l);
 
-	ofono_debug("%s conditions:", cf_type_lut[cf->query_next]);
+	DBG("%s conditions:", cf_type_lut[cf->query_next]);
 	cf_cond_list_print(l);
 
 	if (cf->query_next != cf->query_end) {
 		cf->query_next++;
-		set_query_next_cf_cond(modem);
+		set_query_next_cf_cond(cf);
 	}
 }
 
-static void set_query_next_cf_cond(struct ofono_modem *modem)
+static void set_query_next_cf_cond(struct ofono_call_forwarding *cf)
 {
-	struct call_forwarding_data *cf = modem->call_forwarding;
-
-	cf->ops->query(modem, cf->query_next, BEARER_CLASS_DEFAULT,
-			set_query_cf_callback, modem);
+	cf->driver->query(cf, cf->query_next, BEARER_CLASS_DEFAULT,
+			set_query_cf_callback, cf);
 }
 
 static void set_property_callback(const struct ofono_error *error, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_forwarding_data *cf = modem->call_forwarding;
+	struct ofono_call_forwarding *cf = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Error occurred during set/erasure");
+		DBG("Error occurred during set/erasure");
 		__ofono_dbus_pending_reply(&cf->pending,
 					__ofono_error_failed(cf->pending));
 		return;
 	}
 
 	/* Successfully set, query the entire set just in case */
-	set_query_next_cf_cond(modem);
+	set_query_next_cf_cond(cf);
 }
 
-static DBusMessage *set_property_request(struct ofono_modem *modem,
+static DBusMessage *set_property_request(struct ofono_call_forwarding *cf,
 						DBusMessage *msg,
 						int type, int cls,
 						struct ofono_phone_number *ph,
 						int timeout)
 {
-	struct call_forwarding_data *cf = modem->call_forwarding;
-
-	if (ph->number[0] != '\0' && cf->ops->registration == NULL)
+	if (ph->number[0] != '\0' && cf->driver->registration == NULL)
 		return __ofono_error_not_implemented(msg);
 
-	if (ph->number[0] == '\0' && cf->ops->erasure == NULL)
+	if (ph->number[0] == '\0' && cf->driver->erasure == NULL)
 		return __ofono_error_not_implemented(msg);
 
 	cf->pending = dbus_message_ref(msg);
 	cf->query_next = type;
 	cf->query_end = type;
 
-	ofono_debug("Farming off request, will be erasure: %d",
-			ph->number[0] == '\0');
+	DBG("Farming off request, will be erasure: %d", ph->number[0] == '\0');
 
 	if (ph->number[0] != '\0')
-		cf->ops->registration(modem, type, cls, ph, timeout,
-					set_property_callback, modem);
+		cf->driver->registration(cf, type, cls, ph, timeout,
+					set_property_callback, cf);
 	else
-		cf->ops->erasure(modem, type, cls, set_property_callback, modem);
+		cf->driver->erasure(cf, type, cls, set_property_callback, cf);
 
 	return NULL;
 }
@@ -602,15 +583,15 @@ static DBusMessage *set_property_request(struct ofono_modem *modem,
 static DBusMessage *cf_set_property(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_forwarding_data *cf = modem->call_forwarding;
+	struct ofono_call_forwarding *cf = data;
 	DBusMessageIter iter;
 	DBusMessageIter var;
 	const char *property;
 	int cls;
 	int type;
 
-	if (cf->pending)
+	if (__ofono_call_forwarding_is_busy(cf) ||
+			__ofono_ussd_is_busy(cf->ussd))
 		return __ofono_error_busy(msg);
 
 	if (!dbus_message_iter_init(msg, &iter))
@@ -630,7 +611,7 @@ static DBusMessage *cf_set_property(DBusConnection *conn, DBusMessage *msg,
 	if (cf_condition_timeout_property(property, &cls)) {
 		dbus_uint16_t timeout;
 		GSList *l;
-		struct ofono_cf_condition *c;
+		struct ofono_call_forwarding_condition *c;
 
 		type = CALL_FORWARDING_TYPE_NO_REPLY;
 
@@ -651,7 +632,7 @@ static DBusMessage *cf_set_property(DBusConnection *conn, DBusMessage *msg,
 
 		c = l->data;
 
-		return set_property_request(modem, msg, type, cls,
+		return set_property_request(cf, msg, type, cls,
 						&c->phone_number, timeout);
 	} else if (cf_condition_enabled_property(cf, property, &type, &cls)) {
 		struct ofono_phone_number ph;
@@ -674,7 +655,7 @@ static DBusMessage *cf_set_property(DBusConnection *conn, DBusMessage *msg,
 
 		timeout = cf_find_timeout(cf->cf_conditions[type], cls);
 
-		return set_property_request(modem, msg, type, cls, &ph,
+		return set_property_request(cf, msg, type, cls, &ph,
 						timeout);
 	}
 
@@ -684,11 +665,10 @@ static DBusMessage *cf_set_property(DBusConnection *conn, DBusMessage *msg,
 static void disable_conditional_callback(const struct ofono_error *error,
 						void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_forwarding_data *cf = modem->call_forwarding;
+	struct ofono_call_forwarding *cf = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Error occurred during conditional erasure");
+		DBG("Error occurred during conditional erasure");
 
 		__ofono_dbus_pending_reply(&cf->pending,
 					__ofono_error_failed(cf->pending));
@@ -698,16 +678,15 @@ static void disable_conditional_callback(const struct ofono_error *error,
 	/* Query the three conditional cf types */
 	cf->query_next = CALL_FORWARDING_TYPE_BUSY;
 	cf->query_end = CALL_FORWARDING_TYPE_NOT_REACHABLE;
-	set_query_next_cf_cond(modem);
+	set_query_next_cf_cond(cf);
 }
 
 static void disable_all_callback(const struct ofono_error *error, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_forwarding_data *cf = modem->call_forwarding;
+	struct ofono_call_forwarding *cf = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Error occurred during erasure of all");
+		DBG("Error occurred during erasure of all");
 
 		__ofono_dbus_pending_reply(&cf->pending,
 					__ofono_error_failed(cf->pending));
@@ -717,23 +696,23 @@ static void disable_all_callback(const struct ofono_error *error, void *data)
 	/* Query all cf types */
 	cf->query_next = CALL_FORWARDING_TYPE_UNCONDITIONAL;
 	cf->query_end = CALL_FORWARDING_TYPE_NOT_REACHABLE;
-	set_query_next_cf_cond(modem);
+	set_query_next_cf_cond(cf);
 }
 
 static DBusMessage *cf_disable_all(DBusConnection *conn, DBusMessage *msg,
 					void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_forwarding_data *cf = modem->call_forwarding;
+	struct ofono_call_forwarding *cf = data;
 	const char *strtype;
 	int type;
 
-	if (cf->pending)
-		return __ofono_error_busy(msg);
-
-	if (!cf->ops->erasure)
+	if (!cf->driver->erasure)
 		return __ofono_error_not_implemented(msg);
 
+	if (__ofono_call_forwarding_is_busy(cf) ||
+			__ofono_ussd_is_busy(cf->ussd))
+		return __ofono_error_busy(msg);
+
 	if (dbus_message_get_args(msg, NULL, DBUS_TYPE_STRING, &strtype,
 					DBUS_TYPE_INVALID) == FALSE)
 		return __ofono_error_invalid_args(msg);
@@ -748,11 +727,11 @@ static DBusMessage *cf_disable_all(DBusConnection *conn, DBusMessage *msg,
 	cf->pending = dbus_message_ref(msg);
 
 	if (type == CALL_FORWARDING_TYPE_ALL)
-		cf->ops->erasure(modem, type, BEARER_CLASS_DEFAULT,
-				disable_all_callback, modem);
+		cf->driver->erasure(cf, type, BEARER_CLASS_DEFAULT,
+				disable_all_callback, cf);
 	else
-		cf->ops->erasure(modem, type, BEARER_CLASS_DEFAULT,
-				disable_conditional_callback, modem);
+		cf->driver->erasure(cf, type, BEARER_CLASS_DEFAULT,
+				disable_conditional_callback, cf);
 
 	return NULL;
 }
@@ -772,10 +751,9 @@ static GDBusSignalTable cf_signals[] = {
 	{ }
 };
 
-static DBusMessage *cf_ss_control_reply(struct ofono_modem *modem,
+static DBusMessage *cf_ss_control_reply(struct ofono_call_forwarding *cf,
 					struct cf_ss_request *req)
 {
-	struct call_forwarding_data *cf = modem->call_forwarding;
 	const char *context = "CallForwarding";
 	const char *sig = "(ssa{sv})";
 	const char *ss_type = ss_control_type_to_string(req->ss_type);
@@ -846,11 +824,10 @@ static DBusMessage *cf_ss_control_reply(struct ofono_modem *modem,
 }
 
 static void ss_set_query_cf_callback(const struct ofono_error *error, int total,
-					const struct ofono_cf_condition *list,
-					void *data)
+			const struct ofono_call_forwarding_condition *list,
+			void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_forwarding_data *cf = modem->call_forwarding;
+	struct ofono_call_forwarding *cf = data;
 	GSList *l;
 	DBusMessage *reply;
 
@@ -863,41 +840,38 @@ static void ss_set_query_cf_callback(const struct ofono_error *error, int total,
 	}
 
 	l = cf_cond_list_create(total, list);
-	ofono_debug("%s conditions:", cf_type_lut[cf->query_next]);
+	DBG("%s conditions:", cf_type_lut[cf->query_next]);
 	cf_cond_list_print(l);
 
 	cf->ss_req->cf_list[cf->query_next] = l;
 
 	if (cf->query_next == cf->query_end) {
-		reply = cf_ss_control_reply(modem, cf->ss_req);
+		reply = cf_ss_control_reply(cf, cf->ss_req);
 		__ofono_dbus_pending_reply(&cf->pending, reply);
 		g_free(cf->ss_req);
 		cf->ss_req = NULL;
 	}
 
-	set_new_cond_list(modem, cf->query_next, l);
+	set_new_cond_list(cf, cf->query_next, l);
 
 	if (cf->query_next != cf->query_end) {
 		cf->query_next++;
-		ss_set_query_next_cf_cond(modem);
+		ss_set_query_next_cf_cond(cf);
 	}
 }
 
-static void ss_set_query_next_cf_cond(struct ofono_modem *modem)
+static void ss_set_query_next_cf_cond(struct ofono_call_forwarding *cf)
 {
-	struct call_forwarding_data *cf = modem->call_forwarding;
-
-	cf->ops->query(modem, cf->query_next, BEARER_CLASS_DEFAULT,
-			ss_set_query_cf_callback, modem);
+	cf->driver->query(cf, cf->query_next, BEARER_CLASS_DEFAULT,
+			ss_set_query_cf_callback, cf);
 }
 
 static void cf_ss_control_callback(const struct ofono_error *error, void *data)
 {
-	struct ofono_modem *modem = data;
-	struct call_forwarding_data *cf = modem->call_forwarding;
+	struct ofono_call_forwarding *cf = data;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Error occurred during cf ss control set/erasure");
+		DBG("Error occurred during cf ss control set/erasure");
 
 		__ofono_dbus_pending_reply(&cf->pending,
 					__ofono_error_failed(cf->pending));
@@ -906,16 +880,15 @@ static void cf_ss_control_callback(const struct ofono_error *error, void *data)
 		return;
 	}
 
-	ss_set_query_next_cf_cond(modem);
+	ss_set_query_next_cf_cond(cf);
 }
 
-static gboolean cf_ss_control(struct ofono_modem *modem,
-				enum ss_control_type type, const char *sc,
+static gboolean cf_ss_control(int type, const char *sc,
 				const char *sia, const char *sib,
 				const char *sic, const char *dn,
-				DBusMessage *msg)
+				DBusMessage *msg, void *data)
 {
-	struct call_forwarding_data *cf = modem->call_forwarding;
+	struct ofono_call_forwarding *cf = data;
 	DBusConnection *conn = ofono_dbus_get_connection();
 	int cls = BEARER_CLASS_SS_DEFAULT;
 	int timeout = DEFAULT_NO_REPLY_TIMEOUT;
@@ -928,17 +901,17 @@ static gboolean cf_ss_control(struct ofono_modem *modem,
 	if (!cf)
 		return FALSE;
 
-	if (cf->pending) {
+	if (__ofono_call_forwarding_is_busy(cf)) {
 		reply = __ofono_error_busy(msg);
 		g_dbus_send_message(conn, reply);
 
 		return TRUE;
 	}
 
-	ofono_debug("Received call forwarding ss control request");
+	DBG("Received call forwarding ss control request");
 
-	ofono_debug("type: %d, sc: %s, sia: %s, sib: %s, sic: %s, dn: %s",
-			type, sc, sia, sib, sic, dn);
+	DBG("type: %d, sc: %s, sia: %s, sib: %s, sic: %s, dn: %s",
+		type, sc, sia, sib, sic, dn);
 
 	if (!strcmp(sc, "21"))
 		cf_type = CALL_FORWARDING_TYPE_UNCONDITIONAL;
@@ -961,7 +934,8 @@ static gboolean cf_ss_control(struct ofono_modem *modem,
 		type == SS_CONTROL_TYPE_DEACTIVATION))
 		goto error;
 
-	/* Activation / Registration is figured context specific according to
+	/*
+	 * Activation / Registration is figured context specific according to
 	 * 22.030 Section 6.5.2 "The UE shall determine from the context
 	 * whether, an entry of a single *, activation or registration
 	 * was intended."
@@ -1010,19 +984,19 @@ static gboolean cf_ss_control(struct ofono_modem *modem,
 
 	switch (type) {
 	case SS_CONTROL_TYPE_REGISTRATION:
-		operation = cf->ops->registration;
+		operation = cf->driver->registration;
 		break;
 	case SS_CONTROL_TYPE_ACTIVATION:
-		operation = cf->ops->activation;
+		operation = cf->driver->activation;
 		break;
 	case SS_CONTROL_TYPE_DEACTIVATION:
-		operation = cf->ops->deactivation;
+		operation = cf->driver->deactivation;
 		break;
 	case SS_CONTROL_TYPE_ERASURE:
-		operation = cf->ops->erasure;
+		operation = cf->driver->erasure;
 		break;
 	case SS_CONTROL_TYPE_QUERY:
-		operation = cf->ops->query;
+		operation = cf->driver->query;
 		break;
 	}
 
@@ -1063,7 +1037,8 @@ static gboolean cf_ss_control(struct ofono_modem *modem,
 		break;
 	}
 
-	/* Some modems don't understand all classes very well, particularly
+	/*
+	 * Some modems don't understand all classes very well, particularly
 	 * the older models.  So if the bearer class is the default, we
 	 * just use the more commonly understood value of 7 since BEARER_SMS
 	 * is not applicable to CallForwarding conditions according to 22.004
@@ -1075,23 +1050,23 @@ static gboolean cf_ss_control(struct ofono_modem *modem,
 	switch (cf->ss_req->ss_type) {
 	case SS_CONTROL_TYPE_REGISTRATION:
 		string_to_phone_number(sia, &ph);
-		cf->ops->registration(modem, cf_type, cls, &ph, timeout,
-					cf_ss_control_callback, modem);
+		cf->driver->registration(cf, cf_type, cls, &ph, timeout,
+					cf_ss_control_callback, cf);
 		break;
 	case SS_CONTROL_TYPE_ACTIVATION:
-		cf->ops->activation(modem, cf_type, cls, cf_ss_control_callback,
-					modem);
+		cf->driver->activation(cf, cf_type, cls, cf_ss_control_callback,
+					cf);
 		break;
 	case SS_CONTROL_TYPE_DEACTIVATION:
-		cf->ops->deactivation(modem, cf_type, cls,
-					cf_ss_control_callback, modem);
+		cf->driver->deactivation(cf, cf_type, cls,
+					cf_ss_control_callback, cf);
 		break;
 	case SS_CONTROL_TYPE_ERASURE:
-		cf->ops->erasure(modem, cf_type, cls, cf_ss_control_callback,
-					modem);
+		cf->driver->erasure(cf, cf_type, cls, cf_ss_control_callback,
+					cf);
 		break;
 	case SS_CONTROL_TYPE_QUERY:
-		ss_set_query_next_cf_cond(modem);
+		ss_set_query_next_cf_cond(cf);
 		break;
 	}
 
@@ -1103,79 +1078,179 @@ error:
 	return TRUE;
 }
 
-static void cf_register_ss_controls(struct ofono_modem *modem)
+static void cf_register_ss_controls(struct ofono_call_forwarding *cf)
 {
-	ss_control_register(modem, "21", cf_ss_control);
-	ss_control_register(modem, "67", cf_ss_control);
-	ss_control_register(modem, "61", cf_ss_control);
-	ss_control_register(modem, "62", cf_ss_control);
+	__ofono_ussd_ssc_register(cf->ussd, "21", cf_ss_control, cf, NULL);
+	__ofono_ussd_ssc_register(cf->ussd, "67", cf_ss_control, cf, NULL);
+	__ofono_ussd_ssc_register(cf->ussd, "61", cf_ss_control, cf, NULL);
+	__ofono_ussd_ssc_register(cf->ussd, "62", cf_ss_control, cf, NULL);
 
-	ss_control_register(modem, "002", cf_ss_control);
-	ss_control_register(modem, "004", cf_ss_control);
+	__ofono_ussd_ssc_register(cf->ussd, "002", cf_ss_control, cf, NULL);
+	__ofono_ussd_ssc_register(cf->ussd, "004", cf_ss_control, cf, NULL);
 }
 
-static void cf_unregister_ss_controls(struct ofono_modem *modem)
+static void cf_unregister_ss_controls(struct ofono_call_forwarding *cf)
 {
-	ss_control_unregister(modem, "21", cf_ss_control);
-	ss_control_unregister(modem, "67", cf_ss_control);
-	ss_control_unregister(modem, "61", cf_ss_control);
-	ss_control_unregister(modem, "62", cf_ss_control);
+	__ofono_ussd_ssc_unregister(cf->ussd, "21");
+	__ofono_ussd_ssc_unregister(cf->ussd, "67");
+	__ofono_ussd_ssc_unregister(cf->ussd, "61");
+	__ofono_ussd_ssc_unregister(cf->ussd, "62");
 
-	ss_control_unregister(modem, "002", cf_ss_control);
-	ss_control_unregister(modem, "004", cf_ss_control);
+	__ofono_ussd_ssc_unregister(cf->ussd, "002");
+	__ofono_ussd_ssc_unregister(cf->ussd, "004");
 }
 
-int ofono_call_forwarding_register(struct ofono_modem *modem,
-				struct ofono_call_forwarding_ops *ops)
+gboolean __ofono_call_forwarding_is_busy(struct ofono_call_forwarding *cf)
 {
-	DBusConnection *conn = ofono_dbus_get_connection();
+	return cf->pending ? TRUE : FALSE;
+}
 
-	if (modem == NULL)
-		return -1;
+int ofono_call_forwarding_driver_register(const struct ofono_call_forwarding_driver *d)
+{
+	DBG("driver: %p, name: %s", d, d->name);
 
-	if (ops == NULL)
-		return -1;
+	if (d->probe == NULL)
+		return -EINVAL;
 
-	if (ops->query == NULL)
-		return -1;
-
-	modem->call_forwarding = call_forwarding_create();
-
-	if (modem->call_forwarding == NULL)
-		return -1;
-
-	modem->call_forwarding->ops = ops;
-
-	if (!g_dbus_register_interface(conn, modem->path,
-					CALL_FORWARDING_INTERFACE,
-					cf_methods, cf_signals, NULL,
-					modem, call_forwarding_destroy)) {
-		ofono_error("Could not register CallForwarding %s", modem->path);
-		call_forwarding_destroy(modem);
-
-		return -1;
-	}
-
-	ofono_debug("Registered call forwarding interface");
-
-	cf_register_ss_controls(modem);
-
-	ofono_modem_add_interface(modem, CALL_FORWARDING_INTERFACE);
+	g_drivers = g_slist_prepend(g_drivers, (void *) d);
 
 	return 0;
 }
 
-void ofono_call_forwarding_unregister(struct ofono_modem *modem)
+void ofono_call_forwarding_driver_unregister(const struct ofono_call_forwarding_driver *d)
 {
-	struct call_forwarding_data *cf = modem->call_forwarding;
-	DBusConnection *conn = ofono_dbus_get_connection();
+	DBG("driver: %p, name: %s", d, d->name);
 
-	if (!cf)
+	g_drivers = g_slist_remove(g_drivers, (void *) d);
+}
+
+static void call_forwarding_unregister(struct ofono_atom *atom)
+{
+	struct ofono_call_forwarding *cf = __ofono_atom_get_data(atom);
+	const char *path = __ofono_atom_get_path(cf->atom);
+	DBusConnection *conn = ofono_dbus_get_connection();
+	struct ofono_modem *modem = __ofono_atom_get_modem(cf->atom);
+
+	ofono_modem_remove_interface(modem, OFONO_CALL_FORWARDING_INTERFACE);
+	g_dbus_unregister_interface(conn, path,
+					OFONO_CALL_FORWARDING_INTERFACE);
+
+	if (cf->ussd)
+		cf_unregister_ss_controls(cf);
+
+	if (cf->ussd_watch)
+		__ofono_modem_remove_atom_watch(modem, cf->ussd_watch);
+}
+
+static void call_forwarding_remove(struct ofono_atom *atom)
+{
+	struct ofono_call_forwarding *cf = __ofono_atom_get_data(atom);
+
+	DBG("atom: %p", atom);
+
+	if (cf == NULL)
 		return;
 
-	ofono_modem_remove_interface(modem, CALL_FORWARDING_INTERFACE);
-	g_dbus_unregister_interface(conn, modem->path,
-					CALL_FORWARDING_INTERFACE);
+	if (cf->driver && cf->driver->remove)
+		cf->driver->remove(cf);
 
-	modem->call_forwarding = NULL;
+	cf_clear_all(cf);
+
+	g_free(cf);
+}
+
+struct ofono_call_forwarding *ofono_call_forwarding_create(struct ofono_modem *modem,
+							unsigned int vendor,
+							const char *driver,
+							void *data)
+{
+	struct ofono_call_forwarding *cf;
+	GSList *l;
+
+	if (driver == NULL)
+		return NULL;
+
+	cf = g_try_new0(struct ofono_call_forwarding, 1);
+
+	if (cf == NULL)
+		return NULL;
+
+	cf->atom = __ofono_modem_add_atom(modem,
+						OFONO_ATOM_TYPE_CALL_FORWARDING,
+						call_forwarding_remove, cf);
+	for (l = g_drivers; l; l = l->next) {
+		const struct ofono_call_forwarding_driver *drv = l->data;
+
+		if (g_strcmp0(drv->name, driver))
+			continue;
+
+		if (drv->probe(cf, vendor, data) < 0)
+			continue;
+
+		cf->driver = drv;
+		break;
+	}
+
+	return cf;
+}
+
+static void ussd_watch(struct ofono_atom *atom,
+			enum ofono_atom_watch_condition cond, void *data)
+{
+	struct ofono_call_forwarding *cf = data;
+
+	if (cond == OFONO_ATOM_WATCH_CONDITION_UNREGISTERED) {
+		cf->ussd = NULL;
+		return;
+	}
+
+	cf->ussd = __ofono_atom_get_data(atom);
+	cf_register_ss_controls(cf);
+}
+
+void ofono_call_forwarding_register(struct ofono_call_forwarding *cf)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+	const char *path = __ofono_atom_get_path(cf->atom);
+	struct ofono_modem *modem = __ofono_atom_get_modem(cf->atom);
+	struct ofono_atom *ussd_atom;
+
+	if (!g_dbus_register_interface(conn, path,
+					OFONO_CALL_FORWARDING_INTERFACE,
+					cf_methods, cf_signals, NULL, cf,
+					NULL)) {
+		ofono_error("Could not create %s interface",
+				OFONO_CALL_FORWARDING_INTERFACE);
+
+		return;
+	}
+
+	ofono_modem_add_interface(modem, OFONO_CALL_FORWARDING_INTERFACE);
+
+	cf->ussd_watch = __ofono_modem_add_atom_watch(modem,
+					OFONO_ATOM_TYPE_USSD,
+					ussd_watch, cf, NULL);
+
+	ussd_atom = __ofono_modem_find_atom(modem, OFONO_ATOM_TYPE_USSD);
+
+	if (ussd_atom && __ofono_atom_get_registered(ussd_atom))
+		ussd_watch(ussd_atom, OFONO_ATOM_WATCH_CONDITION_REGISTERED,
+				cf);
+
+	__ofono_atom_register(cf->atom, call_forwarding_unregister);
+}
+
+void ofono_call_forwarding_remove(struct ofono_call_forwarding *cf)
+{
+	__ofono_atom_free(cf->atom);
+}
+
+void ofono_call_forwarding_set_data(struct ofono_call_forwarding *cf, void *data)
+{
+	cf->driver_data = data;
+}
+
+void *ofono_call_forwarding_get_data(struct ofono_call_forwarding *cf)
+{
+	return cf->driver_data;
 }
