@@ -57,6 +57,7 @@ DBusMessage *__ofono_error_in_use(DBusMessage *msg);
 DBusMessage *__ofono_error_not_attached(DBusMessage *msg);
 DBusMessage *__ofono_error_attach_in_progress(DBusMessage *msg);
 DBusMessage *__ofono_error_canceled(DBusMessage *msg);
+DBusMessage *__ofono_error_access_denied(DBusMessage *msg);
 
 void __ofono_dbus_pending_reply(DBusMessage **msg, DBusMessage *reply);
 
@@ -89,7 +90,8 @@ void __ofono_plugin_cleanup(void);
 
 #include <ofono/modem.h>
 
-typedef void (*ofono_modem_foreach_func)(struct ofono_modem *modem, void *data);
+typedef void (*ofono_modem_foreach_func)(struct ofono_modem *modem,
+						void *data);
 void __ofono_modem_foreach(ofono_modem_foreach_func cb, void *userdata);
 
 unsigned int __ofono_modem_callid_next(struct ofono_modem *modem);
@@ -120,8 +122,9 @@ enum ofono_atom_type {
 	OFONO_ATOM_TYPE_GPRS = 16,
 	OFONO_ATOM_TYPE_GPRS_CONTEXT = 17,
 	OFONO_ATOM_TYPE_RADIO_SETTINGS = 18,
-	OFONO_ATOM_TYPE_STK = 19,
-	OFONO_ATOM_TYPE_NETTIME = 20,
+	OFONO_ATOM_TYPE_AUDIO_SETTINGS = 19,
+	OFONO_ATOM_TYPE_STK = 20,
+	OFONO_ATOM_TYPE_NETTIME = 21,
 };
 
 enum ofono_atom_watch_condition {
@@ -160,12 +163,28 @@ gboolean __ofono_atom_get_registered(struct ofono_atom *atom);
 unsigned int __ofono_modem_add_atom_watch(struct ofono_modem *modem,
 					enum ofono_atom_type type,
 					ofono_atom_watch_func notify,
-					void *data, ofono_destroy_func destroy);
+					void *data,
+					ofono_destroy_func destroy);
 gboolean __ofono_modem_remove_atom_watch(struct ofono_modem *modem,
 						unsigned int id);
 
 void __ofono_atom_free(struct ofono_atom *atom);
 
+typedef void (*ofono_modemwatch_cb_t)(struct ofono_modem *modem,
+					gboolean added, void *data);
+void __ofono_modemwatch_init();
+void __ofono_modemwatch_cleanup();
+unsigned int __ofono_modemwatch_add(ofono_modemwatch_cb_t cb, void *user,
+					ofono_destroy_func destroy);
+gboolean __ofono_modemwatch_remove(unsigned int id);
+
+typedef void (*ofono_modem_online_notify_func)(ofono_bool_t online, void *data);
+unsigned int __ofono_modem_add_online_watch(struct ofono_modem *modem,
+					ofono_modem_online_notify_func notify,
+					void *data, ofono_destroy_func destroy);
+void __ofono_modem_remove_online_watch(struct ofono_modem *modem,
+					unsigned int id);
+
 #include <ofono/call-barring.h>
 
 gboolean __ofono_call_barring_is_busy(struct ofono_call_barring *cb);
@@ -185,6 +204,7 @@ gboolean __ofono_call_settings_is_busy(struct ofono_call_settings *cs);
 #include <ofono/gprs.h>
 #include <ofono/gprs-context.h>
 #include <ofono/radio-settings.h>
+#include <ofono/audio-settings.h>
 
 #include <ofono/voicecall.h>
 
@@ -195,6 +215,7 @@ enum ofono_voicecall_interaction {
 };
 
 typedef void (*ofono_voicecall_dial_cb_t)(struct ofono_call *call, void *data);
+typedef void (*ofono_voicecall_tone_cb_t)(int error, void *data);
 
 ofono_bool_t __ofono_voicecall_is_busy(struct ofono_voicecall *vc,
 					enum ofono_voicecall_interaction type);
@@ -206,6 +227,11 @@ int __ofono_voicecall_dial(struct ofono_voicecall *vc,
 				ofono_voicecall_dial_cb_t cb, void *user_data);
 void __ofono_voicecall_dial_cancel(struct ofono_voicecall *vc);
 
+int __ofono_voicecall_tone_send(struct ofono_voicecall *vc,
+				const char *tone_str,
+				ofono_voicecall_tone_cb_t cb, void *user_data);
+void __ofono_voicecall_tone_cancel(struct ofono_voicecall *vc, int id);
+
 #include <ofono/sms.h>
 
 struct sms;
@@ -214,21 +240,73 @@ enum ofono_sms_submit_flag {
 	OFONO_SMS_SUBMIT_FLAG_REQUEST_SR =	0x1,
 	OFONO_SMS_SUBMIT_FLAG_RECORD_HISTORY =	0x2,
 	OFONO_SMS_SUBMIT_FLAG_RETRY =		0x4,
+	OFONO_SMS_SUBMIT_FLAG_EXPOSE_DBUS =	0x8,
 };
 
 typedef void (*ofono_sms_txq_submit_cb_t)(gboolean ok, void *data);
+typedef void (*ofono_sms_txq_queued_cb_t)(struct ofono_sms *sms,
+						const struct ofono_uuid *uuid,
+						void *data);
+typedef void (*ofono_sms_text_notify_cb_t)(const char *from,
+						const struct tm *remote,
+						const struct tm *local,
+						const char *text,
+						void *data);
+typedef void (*ofono_sms_datagram_notify_cb_t)(const char *from,
+						const struct tm *remote,
+						const struct tm *local,
+						int dst, int src,
+						const unsigned char *buffer,
+						unsigned int len,
+						void *data);
 
-unsigned int __ofono_sms_txq_submit(struct ofono_sms *sms, GSList *list,
-					unsigned int flags,
+int __ofono_sms_txq_submit(struct ofono_sms *sms, GSList *list,
+				unsigned int flags, struct ofono_uuid *uuid,
+				ofono_sms_txq_queued_cb_t, void *data);
+
+int __ofono_sms_txq_set_submit_notify(struct ofono_sms *sms,
+					struct ofono_uuid *uuid,
 					ofono_sms_txq_submit_cb_t cb,
+					void *data,
+					ofono_destroy_func destroy);
+
+const char *__ofono_sms_message_path_from_uuid(struct ofono_sms *sms,
+						const struct ofono_uuid *uuid);
+
+unsigned int __ofono_sms_text_watch_add(struct ofono_sms *sms,
+					ofono_sms_text_notify_cb_t cb,
 					void *data, ofono_destroy_func destroy);
+gboolean __ofono_sms_text_watch_remove(struct ofono_sms *sms,
+					unsigned int id);
+
+unsigned int __ofono_sms_datagram_watch_add(struct ofono_sms *sms,
+					ofono_sms_datagram_notify_cb_t cb,
+					int dst, int src, void *data,
+					ofono_destroy_func destroy);
+gboolean __ofono_sms_datagram_watch_remove(struct ofono_sms *sms,
+					unsigned int id);
+
+unsigned short __ofono_sms_get_next_ref(struct ofono_sms *sms);
 
 #include <ofono/sim.h>
+
+ofono_bool_t __ofono_sim_service_available(struct ofono_sim *sim,
+						int ust_service,
+						int sst_service);
+
 #include <ofono/stk.h>
 
+typedef void (*__ofono_sms_sim_download_cb_t)(ofono_bool_t ok,
+						const unsigned char *tp_ud,
+						int len, void *data);
+
 struct cbs;
 void __ofono_cbs_sim_download(struct ofono_stk *stk, const struct cbs *msg);
 
+struct sms;
+int __ofono_sms_sim_download(struct ofono_stk *stk, const struct sms *msg,
+				__ofono_sms_sim_download_cb_t cb, void *data);
+
 #include <ofono/ssn.h>
 
 typedef void (*ofono_ssn_mo_notify_cb)(int index, void *user);
@@ -239,12 +317,12 @@ typedef void (*ofono_ssn_mt_notify_cb)(int index,
 unsigned int __ofono_ssn_mo_watch_add(struct ofono_ssn *ssn, int code1,
 					ofono_ssn_mo_notify_cb cb, void *user,
 					ofono_destroy_func destroy);
-gboolean __ofono_ssn_mo_watch_remove(struct ofono_ssn *ssn, int id);
+gboolean __ofono_ssn_mo_watch_remove(struct ofono_ssn *ssn, unsigned int id);
 
 unsigned int __ofono_ssn_mt_watch_add(struct ofono_ssn *ssn, int code2,
 					ofono_ssn_mt_notify_cb cb, void *user,
 					ofono_destroy_func destroy);
-gboolean __ofono_ssn_mt_watch_remove(struct ofono_ssn *ssn, int id);
+gboolean __ofono_ssn_mt_watch_remove(struct ofono_ssn *ssn, unsigned int id);
 
 #include <ofono/ussd.h>
 
@@ -258,6 +336,10 @@ typedef gboolean (*ofono_ussd_passwd_cb_t)(const char *sc,
 					const char *old, const char *new,
 					DBusMessage *msg, void *data);
 
+typedef void (*ofono_ussd_request_cb_t)(int error, int dcs,
+					const unsigned char *pdu, int len,
+					void *data);
+
 gboolean __ofono_ussd_ssc_register(struct ofono_ussd *ussd, const char *sc,
 					ofono_ussd_ssc_cb_t cb, void *data,
 					ofono_destroy_func destroy);
@@ -269,6 +351,11 @@ gboolean __ofono_ussd_passwd_register(struct ofono_ussd *ussd, const char *sc,
 void __ofono_ussd_passwd_unregister(struct ofono_ussd *ussd, const char *sc);
 gboolean __ofono_ussd_is_busy(struct ofono_ussd *ussd);
 
+int __ofono_ussd_initiate(struct ofono_ussd *ussd, int dcs,
+			const unsigned char *pdu, int len,
+			ofono_ussd_request_cb_t cb, void *user_data);
+void __ofono_ussd_initiate_cancel(struct ofono_ussd *ussd);
+
 #include <ofono/netreg.h>
 
 typedef void (*ofono_netreg_status_notify_cb_t)(int status, int lac, int ci,
@@ -297,17 +384,20 @@ void __ofono_history_call_missed(struct ofono_modem *modem,
 				const struct ofono_call *call, time_t when);
 
 void __ofono_history_sms_received(struct ofono_modem *modem,
-					unsigned int msg_id, const char *from,
+					const struct ofono_uuid *uuid,
+					const char *from,
 					const struct tm *remote,
 					const struct tm *local,
 					const char *text);
 
 void __ofono_history_sms_send_pending(struct ofono_modem *modem,
-					unsigned int msg_id, const char *to,
+					const struct ofono_uuid *uuid,
+					const char *to,
 					time_t when, const char *text);
 
 void __ofono_history_sms_send_status(struct ofono_modem *modem,
-					unsigned int msg_id, time_t when,
+					const struct ofono_uuid *uuid,
+					time_t when,
 					enum ofono_history_sms_status status);
 
 #include <ofono/message-waiting.h>
