@@ -47,6 +47,8 @@
 #define SIM_CACHE_PATH SIM_CACHE_BASEPATH "/%04x"
 #define SIM_CACHE_HEADER_SIZE 38
 #define SIM_FILE_INFO_SIZE 6
+#define SIM_IMAGE_CACHE_BASEPATH STORAGEDIR "/%s-%i/images"
+#define SIM_IMAGE_CACHE_PATH SIM_IMAGE_CACHE_BASEPATH "/%d.xpm"
 
 #define SIM_FS_VERSION 1
 
@@ -56,8 +58,10 @@ static gboolean sim_fs_op_read_block(gpointer user_data);
 
 struct sim_fs_op {
 	int id;
+	unsigned char *buffer;
 	enum ofono_sim_file_structure structure;
 	unsigned short offset;
+	gboolean info_only;
 	int num_bytes;
 	int length;
 	int record_length;
@@ -69,6 +73,7 @@ struct sim_fs_op {
 
 static void sim_fs_op_free(struct sim_fs_op *node)
 {
+	g_free(node->buffer);
 	g_free(node);
 }
 
@@ -77,7 +82,6 @@ struct sim_fs {
 	gint op_source;
 	unsigned char bitmap[32];
 	int fd;
-	unsigned char *buffer;
 	struct ofono_sim *sim;
 	const struct ofono_sim_driver *driver;
 };
@@ -94,7 +98,7 @@ void sim_fs_free(struct sim_fs *fs)
 	 * for operations still in progress
 	 */
 	if (fs->op_q) {
-		g_queue_foreach(fs->op_q, (GFunc)sim_fs_op_free, NULL);
+		g_queue_foreach(fs->op_q, (GFunc) sim_fs_op_free, NULL);
 		g_queue_free(fs->op_q);
 	}
 
@@ -129,9 +133,6 @@ static void sim_fs_end_current(struct sim_fs *fs)
 		fs->fd = -1;
 	}
 
-	g_free(fs->buffer);
-	fs->buffer = NULL;
-
 	memset(fs->bitmap, 0, sizeof(fs->bitmap));
 
 	sim_fs_op_free(op);
@@ -141,7 +142,10 @@ static void sim_fs_op_error(struct sim_fs *fs)
 {
 	struct sim_fs_op *op = g_queue_peek_head(fs->op_q);
 
-	if (op->is_read == TRUE)
+	if (op->info_only == TRUE)
+		((sim_fs_read_info_cb_t) op->cb)
+			(0, 0, 0, 0, op->userdata);
+	else if (op->is_read == TRUE)
 		((ofono_sim_file_read_cb_t) op->cb)
 			(0, 0, 0, 0, 0, op->userdata);
 	else
@@ -228,15 +232,19 @@ static void sim_fs_op_read_block_cb(const struct ofono_error *error,
 	if (op->current == start_block) {
 		bufoff = 0;
 		dataoff = op->offset % 256;
-		tocopy = MIN(256 - op->offset % 256, len);
+		tocopy = MIN(256 - op->offset % 256,
+				op->num_bytes - op->current * 256);
 	} else {
 		bufoff = (op->current - start_block - 1) * 256 +
 				op->offset % 256;
 		dataoff = 0;
-		tocopy = len;
+		tocopy = MIN(256, op->num_bytes - op->current * 256);
 	}
 
-	memcpy(fs->buffer + bufoff, data + dataoff, tocopy);
+	DBG("bufoff: %d, dataoff: %d, tocopy: %d",
+				bufoff, dataoff, tocopy);
+
+	memcpy(op->buffer + bufoff, data + dataoff, tocopy);
 	cache_block(fs, op->current, 256, data, len);
 
 	op->current++;
@@ -244,7 +252,7 @@ static void sim_fs_op_read_block_cb(const struct ofono_error *error,
 	if (op->current > end_block) {
 		ofono_sim_file_read_cb_t cb = op->cb;
 
-		cb(1, op->num_bytes, 0, fs->buffer,
+		cb(1, op->num_bytes, 0, op->buffer,
 				op->record_length, op->userdata);
 
 		sim_fs_end_current(fs);
@@ -265,9 +273,9 @@ static gboolean sim_fs_op_read_block(gpointer user_data)
 	end_block = (op->offset + (op->num_bytes - 1)) / 256;
 
 	if (op->current == start_block) {
-		fs->buffer = g_try_new0(unsigned char, op->num_bytes);
+		op->buffer = g_try_new0(unsigned char, op->num_bytes);
 
-		if (fs->buffer == NULL) {
+		if (op->buffer == NULL) {
 			sim_fs_op_error(fs);
 			return FALSE;
 		}
@@ -288,18 +296,21 @@ static gboolean sim_fs_op_read_block(gpointer user_data)
 			seekoff = SIM_CACHE_HEADER_SIZE + op->current * 256 +
 				op->offset % 256;
 			toread = MIN(256 - op->offset % 256,
-					op->length - op->current * 256);
+					op->num_bytes - op->current * 256);
 		} else {
 			bufoff = (op->current - start_block - 1) * 256 +
 					op->offset % 256;
 			seekoff = SIM_CACHE_HEADER_SIZE + op->current * 256;
-			toread = MIN(256, op->length - op->current * 256);
+			toread = MIN(256, op->num_bytes - op->current * 256);
 		}
 
+		DBG("bufoff: %d, seekoff: %d, toread: %d",
+				bufoff, seekoff, toread);
+
 		if (lseek(fs->fd, seekoff, SEEK_SET) == (off_t) -1)
 			break;
 
-		if (TFR(read(fs->fd, fs->buffer + bufoff, toread)) != toread)
+		if (TFR(read(fs->fd, op->buffer + bufoff, toread)) != toread)
 			break;
 
 		op->current += 1;
@@ -308,7 +319,7 @@ static gboolean sim_fs_op_read_block(gpointer user_data)
 	if (op->current > end_block) {
 		ofono_sim_file_read_cb_t cb = op->cb;
 
-		cb(1, op->num_bytes, 0, fs->buffer,
+		cb(1, op->num_bytes, 0, op->buffer,
 				op->record_length, op->userdata);
 
 		sim_fs_end_current(fs);
@@ -426,7 +437,9 @@ static gboolean sim_fs_op_read_record(gpointer user)
 static void sim_fs_op_info_cb(const struct ofono_error *error, int length,
 				enum ofono_sim_file_structure structure,
 				int record_length,
-				const unsigned char access[3], void *data)
+				const unsigned char access[3],
+				unsigned char file_status,
+				void *data)
 {
 	struct sim_fs *fs = data;
 	struct sim_fs_op *op = g_queue_peek_head(fs->op_q);
@@ -473,14 +486,33 @@ static void sim_fs_op_info_cb(const struct ofono_error *error, int length,
 
 		op->record_length = length;
 		op->current = op->offset / 256;
-		fs->op_source = g_idle_add(sim_fs_op_read_block, fs);
+
+		if (op->info_only == FALSE)
+			fs->op_source = g_idle_add(sim_fs_op_read_block, fs);
 	} else {
 		op->record_length = record_length;
 		op->current = 1;
-		fs->op_source = g_idle_add(sim_fs_op_read_record, fs);
+
+		if (op->info_only == FALSE)
+			fs->op_source = g_idle_add(sim_fs_op_read_record, fs);
 	}
 
-	if (imsi == NULL || cache == FALSE)
+	if (op->info_only == TRUE) {
+		/*
+		 * It's info-only request. So there is no need to request
+		 * actual contents of the EF-files. Just return the EF-info.
+		 */
+		sim_fs_read_info_cb_t cb = op->cb;
+
+		cb(1, file_status, op->length,
+			op->record_length, op->userdata);
+
+		sim_fs_end_current(fs);
+
+		return;
+	}
+
+	if (imsi == NULL || phase == OFONO_SIM_PHASE_UNKNOWN || cache == FALSE)
 		return;
 
 	memset(fileinfo, 0, SIM_CACHE_HEADER_SIZE);
@@ -522,7 +554,7 @@ static gboolean sim_fs_op_check_cached(struct sim_fs *fs)
 	enum ofono_sim_file_structure structure;
 	int record_length;
 
-	if (!imsi)
+	if (imsi == NULL || op->info_only == TRUE)
 		return FALSE;
 
 	path = g_strdup_printf(SIM_CACHE_PATH, imsi, phase, op->id);
@@ -608,17 +640,17 @@ static gboolean sim_fs_op_next(gpointer user_data)
 		switch (op->structure) {
 		case OFONO_SIM_FILE_STRUCTURE_TRANSPARENT:
 			driver->write_file_transparent(fs->sim, op->id, 0,
-					op->length, fs->buffer,
+					op->length, op->buffer,
 					sim_fs_op_write_cb, fs);
 			break;
 		case OFONO_SIM_FILE_STRUCTURE_FIXED:
 			driver->write_file_linear(fs->sim, op->id, op->current,
-					op->length, fs->buffer,
+					op->length, op->buffer,
 					sim_fs_op_write_cb, fs);
 			break;
 		case OFONO_SIM_FILE_STRUCTURE_CYCLIC:
 			driver->write_file_cyclic(fs->sim, op->id,
-					op->length, fs->buffer,
+					op->length, op->buffer,
 					sim_fs_op_write_cb, fs);
 			break;
 		default:
@@ -626,13 +658,50 @@ static gboolean sim_fs_op_next(gpointer user_data)
 					"this can't happen");
 		}
 
-		g_free(fs->buffer);
-		fs->buffer = NULL;
+		g_free(op->buffer);
+		op->buffer = NULL;
 	}
 
 	return FALSE;
 }
 
+int sim_fs_read_info(struct sim_fs *fs, int id,
+			enum ofono_sim_file_structure expected_type,
+			sim_fs_read_info_cb_t cb, void *data)
+{
+	struct sim_fs_op *op;
+
+	if (!cb)
+		return -EINVAL;
+
+	if (!fs->driver)
+		return -EINVAL;
+
+	if (!fs->driver->read_file_info)
+		return -ENOSYS;
+
+	if (!fs->op_q)
+		fs->op_q = g_queue_new();
+
+	op = g_try_new0(struct sim_fs_op, 1);
+	if (op == NULL)
+		return -ENOMEM;
+
+	op->id = id;
+	op->structure = expected_type;
+	op->cb = cb;
+	op->userdata = data;
+	op->is_read = TRUE;
+	op->info_only = TRUE;
+
+	g_queue_push_tail(fs->op_q, op);
+
+	if (g_queue_get_length(fs->op_q) == 1)
+		fs->op_source = g_idle_add(sim_fs_op_next, fs);
+
+	return 0;
+}
+
 int sim_fs_read(struct sim_fs *fs, int id,
 		enum ofono_sim_file_structure expected_type,
 		unsigned short offset, unsigned short num_bytes,
@@ -641,18 +710,20 @@ int sim_fs_read(struct sim_fs *fs, int id,
 	struct sim_fs_op *op;
 
 	if (!cb)
-		return -1;
+		return -EINVAL;
 
 	if (!fs->driver)
-		return -1;
+		return -EINVAL;
 
 	if (!fs->driver->read_file_info)
-		return -1;
+		return -ENOSYS;
 
 	if (!fs->op_q)
 		fs->op_q = g_queue_new();
 
-	op = g_new0(struct sim_fs_op, 1);
+	op = g_try_new0(struct sim_fs_op, 1);
+	if (op == NULL)
+		return -ENOMEM;
 
 	op->id = id;
 	op->structure = expected_type;
@@ -661,6 +732,7 @@ int sim_fs_read(struct sim_fs *fs, int id,
 	op->is_read = TRUE;
 	op->offset = offset;
 	op->num_bytes = num_bytes;
+	op->info_only = FALSE;
 
 	g_queue_push_tail(fs->op_q, op);
 
@@ -678,10 +750,10 @@ int sim_fs_write(struct sim_fs *fs, int id, ofono_sim_file_write_cb_t cb,
 	gconstpointer fn = NULL;
 
 	if (!cb)
-		return -1;
+		return -EINVAL;
 
 	if (!fs->driver)
-		return -1;
+		return -EINVAL;
 
 	switch (structure) {
 	case OFONO_SIM_FILE_STRUCTURE_TRANSPARENT:
@@ -698,18 +770,20 @@ int sim_fs_write(struct sim_fs *fs, int id, ofono_sim_file_write_cb_t cb,
 	}
 
 	if (fn == NULL)
-		return -1;
+		return -ENOSYS;
 
 	if (!fs->op_q)
 		fs->op_q = g_queue_new();
 
-	op = g_new0(struct sim_fs_op, 1);
+	op = g_try_new0(struct sim_fs_op, 1);
+	if (op == NULL)
+		return -ENOMEM;
 
 	op->id = id;
 	op->cb = cb;
 	op->userdata = userdata;
 	op->is_read = FALSE;
-	fs->buffer = g_memdup(data, length);
+	op->buffer = g_memdup(data, length);
 	op->structure = structure;
 	op->length = length;
 	op->current = record;
@@ -722,6 +796,74 @@ int sim_fs_write(struct sim_fs *fs, int id, ofono_sim_file_write_cb_t cb,
 	return 0;
 }
 
+void sim_fs_cache_image(struct sim_fs *fs, const char *image, int id)
+{
+	const char *imsi;
+	enum ofono_sim_phase phase;
+
+	if (fs == NULL || image == NULL)
+		return;
+
+	imsi = ofono_sim_get_imsi(fs->sim);
+	if (imsi == NULL)
+		return;
+
+	phase = ofono_sim_get_phase(fs->sim);
+	if (phase == OFONO_SIM_PHASE_UNKNOWN)
+		return;
+
+	write_file((const unsigned char *) image, strlen(image),
+			SIM_CACHE_MODE, SIM_IMAGE_CACHE_PATH, imsi,
+			phase, id);
+}
+
+char *sim_fs_get_cached_image(struct sim_fs *fs, int id)
+{
+	const char *imsi;
+	enum ofono_sim_phase phase;
+	unsigned short image_length;
+	int fd;
+	char *buffer;
+	char *path;
+	int len;
+	struct stat st_buf;
+
+	if (fs == NULL)
+		return NULL;
+
+	imsi = ofono_sim_get_imsi(fs->sim);
+	if (imsi == NULL)
+		return NULL;
+
+	phase = ofono_sim_get_phase(fs->sim);
+	path = g_strdup_printf(SIM_IMAGE_CACHE_PATH, imsi, phase, id);
+
+	TFR(stat(path, &st_buf));
+	fd = TFR(open(path, O_RDONLY));
+	g_free(path);
+
+	if (fd < 0)
+		return NULL;
+
+	image_length = st_buf.st_size;
+	buffer = g_try_new0(char, image_length + 1);
+
+	if (buffer == NULL) {
+		TFR(close(fd));
+		return NULL;
+	}
+
+	len = TFR(read(fd, buffer, image_length));
+	TFR(close(fd));
+
+	if (len != image_length) {
+		g_free(buffer);
+		return NULL;
+	}
+
+	return buffer;
+}
+
 static void remove_cachefile(const char *imsi, enum ofono_sim_phase phase,
 				const struct dirent *file)
 {
@@ -739,6 +881,23 @@ static void remove_cachefile(const char *imsi, enum ofono_sim_phase phase,
 	g_free(path);
 }
 
+static void remove_imagefile(const char *imsi, enum ofono_sim_phase phase,
+				const struct dirent *file)
+{
+	int id;
+	char *path;
+
+	if (file->d_type != DT_REG)
+		return;
+
+	if (sscanf(file->d_name, "%d", &id) != 1)
+		return;
+
+	path = g_strdup_printf(SIM_IMAGE_CACHE_PATH, imsi, phase, id);
+	remove(path);
+	g_free(path);
+}
+
 void sim_fs_check_version(struct sim_fs *fs)
 {
 	const char *imsi = ofono_sim_get_imsi(fs->sim);
@@ -768,6 +927,20 @@ void sim_fs_check_version(struct sim_fs *fs)
 		g_free(entries);
 	}
 
+	path = g_strdup_printf(SIM_IMAGE_CACHE_BASEPATH, imsi, phase);
+	len = scandir(path, &entries, NULL, alphasort);
+	g_free(path);
+
+	if (len > 0) {
+		/* Remove everything */
+		while (len--) {
+			remove_imagefile(imsi, phase, entries[len]);
+			g_free(entries[len]);
+		}
+
+		g_free(entries);
+	}
+
 	version = SIM_FS_VERSION;
 	write_file(&version, 1, SIM_CACHE_MODE, SIM_CACHE_VERSION, imsi, phase);
 }
